/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _Device
#define _Device

/* Foreward references to bypass cyclic references...
*/
struct DeviceOrientation;
struct DeviceForm;
struct DeviceTray;
struct DeviceMedia;
struct DeviceResolution;
struct DeviceCommand;
struct DeviceData;
struct DevicePrintMode;
struct DeviceInstance;
struct DeviceDither;
struct DeviceBlitter;
struct PluggableBlitter;
struct DeviceGamma;
struct Device;

#include "DeviceOrientation.hpp"
#include "DeviceForm.hpp"
#include "DeviceTray.hpp"
#include "DeviceMedia.hpp"
#include "DeviceResolution.hpp"
#include "DeviceCommand.hpp"
#include "DeviceData.hpp"
#include "DevicePrintMode.hpp"
#include "DeviceInstance.hpp"
#include "DeviceDither.hpp"
#include "GplCompression.hpp"
#include "GplDitherInstance.hpp"
#include "DeviceBlitter.hpp"
#include "PluggableBlitter.hpp"
#include "DeviceGamma.hpp"
#include "Capability.hpp"
#include "PDL.hpp"
#include "DebugOutput.hpp"
#include "StringResource.hpp"

#include <iostream>

// exported by device
extern Device *newDevice (bool fAdvanced);
extern Device *newDevice (char *pszJobProperties, bool fAdvanced);

typedef Device * (*PFNNEWDEVICEWARGS) (char *pszJobProperties, bool fAdvanced);
typedef Device * (*PFNNEWDEVICE)      (bool fAdvanced);
typedef void     (*PFNDELETEDEVICE)   (Device *pDevice);

typedef void     (*PFNOUTPUTFUNCTION) (void *pMagicCookie, unsigned char *puchData, int iSize);

class Device
{
public:
                              Device                    ();
   virtual                   ~Device                    ();

   virtual void               initialize                () = 0;

   virtual char              *getVersion                () = 0;

   virtual char              *getDriverName             () = 0;
   virtual char              *getDeviceName             () = 0;
   virtual char              *getShortName              () = 0;

   virtual DeviceInstance    *getDeviceInstance         () = 0;

   virtual DeviceOrientation *getCurrentOrientation     () = 0;
   virtual char              *getCurrentDitherID        () = 0;
   virtual DeviceDither      *getCurrentDither          () = 0;
   virtual DeviceForm        *getCurrentForm            () = 0;
   virtual DeviceTray        *getCurrentTray            () = 0;
   virtual DeviceMedia       *getCurrentMedia           () = 0;
   virtual DeviceResolution  *getCurrentResolution      () = 0;
   virtual DevicePrintMode   *getCurrentPrintMode       () = 0;
   virtual DeviceCommand     *getCommands               () = 0;
   virtual DeviceData        *getDeviceData             () = 0;
   virtual DeviceGamma       *getCurrentGamma           () = 0;

   virtual int                getPDLLevel               () = 0;
   virtual int                getPDLSubLevel            () = 0;
   virtual int                getPDLMajorRevisionLevel  () = 0;
   virtual int                getPDLMinorRevisionLevel  () = 0;

   virtual bool               hasCapability             (long               lMask) = 0;
   virtual bool               hasRasterCapability       (long               lMask) = 0;
   virtual bool               hasDeviceOption           (long               lMask) = 0;
   virtual bool               hasDeviceOption2          (long               lMask) = 0;

   virtual int                getScanlineMultiple       () = 0;

   virtual bool               beginJob                  () = 0;
   virtual bool               beginJob                  (char              *pszJobProperties) = 0;
   virtual bool               newFrame                  () = 0;
   virtual bool               newFrame                  (char              *pszJobProperties) = 0;
   virtual bool               endJob                    () = 0;
   virtual bool               abortJob                  () = 0;

   virtual bool               rasterize                 (PBYTE              pbBits,
                                                         PBITMAPINFO2       pbmi,
                                                         PRECTL             prectlPageLocation,
                                                         BITBLT_TYPE        eType) = 0;

   virtual void               setOutputStream           (ostream           *osNew) = 0;
   virtual void               setOutputStream           (int                iFileNo) = 0;
   virtual void               setOutputFunction         (PFNOUTPUTFUNCTION  pfnOutputFunction,
                                                         void              *pMagicCookie) = 0;

   virtual bool               sendBinaryDataToDevice    (BinaryData        *pData) = 0;
   virtual bool               sendBinaryDataToDevice    (DeviceForm        *pForm) = 0;
   virtual bool               sendBinaryDataToDevice    (DeviceTray        *pTray) = 0;
   virtual bool               sendBinaryDataToDevice    (DeviceMedia       *pMedia) = 0;
   virtual bool               sendBinaryDataToDevice    (DeviceResolution  *pResolution) = 0;
   virtual bool               sendBinaryDataToDevice    (PBYTE              pbData,
                                                         int                iLength) = 0;

   virtual bool               sendPrintfToDevice        (BinaryData        *pData,
                                                                            ...) = 0;
   virtual bool               sendVPrintfToDevice       (BinaryData        *pData,
                                                         va_list            list) = 0;

   virtual string            *queryCurrentJobProperties () = 0;
   virtual Enumeration       *listJobPropertyKeys       () = 0;
   virtual Enumeration       *listDeviceJobPropertyKeys () = 0;
   virtual string            *getJobPropertyType        (char              *pszKey) = 0;
   virtual string            *getJobProperty            (char              *pszKey) = 0;

   virtual DeviceBlitter     *getDeviceBlitter          () = 0;

   virtual void               loadLibrary               (char              *pszLibraryName) = 0;
   virtual void              *dlsym                     (char              *pszLibraryName,
                                                         char              *pszSymbol) = 0;

   virtual string             toString                  (ostrstream& oss) = 0;
   friend ostream&            operator<<                (ostream& os, const Device& self);
};

#endif
