/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Device.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <fstream.h>

#include <glib.h>
#include <gmodule.h>

// Typedefs to dither prototypes...
typedef DeviceDither * (*PFNCREATEDITHERINSTANCE) (char        *pszDitherType,
                                                   Device      *pDevice,
                                                   char        *pszOptions);

typedef int            (*PFNDITHERNAMEVALID)      (char        *pszId);
typedef bool           (*PFNDITHERCATAGORYVALID)  (char        *pszId);
typedef char         * (*PFNGETDITHERCATAGORY)    (char        *pszId);

typedef Enumeration  * (*PFNGETENUMERATION)      ();

// Function prototypes...
bool          ditherLibraryValid    (char *pszName);
char         *queryLibrary          (char *pszFillIn, char *pszID);
char         *convert               (char *pszFillIn, char *pszPFN, char *pszLibraryName);
char         *truncate              (char *pszLibraryName);

char *
truncate (char *pszLibraryName)
{
   if (  0 == strncmp ("lib", pszLibraryName, 3)                           // has the head part
      && 0 == strcmp (".so", pszLibraryName + strlen (pszLibraryName) - 3) // has the tail part
      )
   {
      pszLibraryName += 3;                                // remove the head part
      pszLibraryName[strlen (pszLibraryName) - 3] = '\0'; // remove the tail part
   }

   return pszLibraryName;
}

char * DeviceDither::
getName (char *pszId)
{
   return StringResource::getString (StringResource::STRINGGROUP_DITHERS,
                                     pszId);
}

char * DeviceDither::
getName ()
{
   return StringResource::getString (StringResource::STRINGGROUP_DITHERS,
                                     getID ());
}

DeviceDither * DeviceDither::
createDitherInstance (char        *pszDitherType,
                      Device      *pDevice,
                      char        *pszOptions)
{
   char  achLibraryName[512]; // @TBD
   char  achFN[512];          // @TBD
   char *pszLibraryName      = 0;

   if (!GplDitherInstance::ditherNameValid (pszDitherType))
      pszLibraryName = queryLibrary (achLibraryName, pszDitherType);

   if (pszLibraryName)
   {
      PFNCREATEDITHERINSTANCE pfnCreateDitherInstance = 0;

      pDevice->loadLibrary (pszLibraryName);

      pfnCreateDitherInstance = (PFNCREATEDITHERINSTANCE)pDevice->dlsym (pszLibraryName,
                                                                         convert (achFN,
                                                                                  "createDitherInstance",
                                                                                  pszLibraryName));

      if (pfnCreateDitherInstance)
      {
         return pfnCreateDitherInstance (pszDitherType,
                                         pDevice,
                                         pszOptions);
      }
      else
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::createDitherInstance: Failed to find createDitherInstance in " << pszLibraryName << endl;
#endif
         return 0;
      }
   }

   return GplDitherInstance::createDitherInstance (pszDitherType,
                                                   pDevice,
                                                   pszOptions);
}

// This is here for consistency's sake... So that the code looks similiar
// to objects which use numeric IDs instead of string IDs.
char * DeviceDither::
IDToName (char *pszId)
{
   return pszId;
}

bool DeviceDither::
ditherNameValid (char *pszId)
{
   char  achLibraryName[512]; // @TBD
   char  achFN[512];          // @TBD
   char *pszLibraryName      = 0;
   bool  fRet;

   fRet = GplDitherInstance::ditherNameValid (pszId);

   if (!fRet)
   {
      pszLibraryName = queryLibrary (achLibraryName, pszId);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherNameValid: queryLibrary (" << pszId << ") = " << pszLibraryName << endl;
#endif

      if (!g_module_supported ())
      {
         cerr << "DeviceDither::ditherNameValid: This program needs glib's module routines!" << endl;

         return false;
      }

      if (pszLibraryName)
      {
         PFNDITHERNAMEVALID  pfnDitherNameValid = 0;
         GModule            *hModLibrary        = 0;

         hModLibrary = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherNameValid: g_module_open (" << pszLibraryName << ") = 0x" << hex << (int)hModLibrary << endl;
#endif

         if (hModLibrary)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherNameValid: dlsym (" << convert (achFN, "ditherNameValid", pszLibraryName) << ")";
#endif

            g_module_symbol (hModLibrary,
                             convert (achFN,
                                      "ditherNameValid",
                                      pszLibraryName),
                             (gpointer *)&pfnDitherNameValid);

#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << " = 0x" << hex << (int)pfnDitherNameValid << endl;
#endif

            if (pfnDitherNameValid)
            {
               fRet = pfnDitherNameValid (pszId);

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherNameValid: it returns " << fRet << endl;
#endif
            }

            g_module_close (hModLibrary);
         }
      }
   }

   return fRet;
}

char * DeviceDither::
getDitherCatagory (char *pszId)
{
   char  achLibraryName[512]; // @TBD
   char  achFN[512];          // @TBD
   char *pszLibraryName      = 0;
   char *pszRet;

   pszRet = GplDitherInstance::getDitherCatagory (pszId);

   if (!pszRet)
   {
      pszLibraryName = queryLibrary (achLibraryName, pszId);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getDitherCatagory: queryLibrary (" << pszId << ") = " << pszLibraryName << endl;
#endif

      if (!g_module_supported ())
      {
         cerr << "DeviceDither::getDitherCatagory: This program needs glib's module routines!" << endl;

         return pszRet;
      }

      if (pszLibraryName)
      {
         PFNGETDITHERCATAGORY  pfnGetDitherCatagory = 0;
         GModule              *hModLibrary          = 0;

         hModLibrary = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getDitherCatagory: g_module_open (" << pszLibraryName << ") = 0x" << hex << (int)hModLibrary << endl;
#endif

         if (hModLibrary)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getDitherCatagory: dlsym (" << convert (achFN, "getDitherCatagory", pszLibraryName) << ")";
#endif

            g_module_symbol (hModLibrary,
                             convert (achFN,
                                      "getDitherCatagory",
                                      pszLibraryName),
                             (gpointer *)&pfnGetDitherCatagory);

#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << " = 0x" << hex << (int)pfnGetDitherCatagory << endl;
#endif

            if (pfnGetDitherCatagory)
            {
               pszRet = pfnGetDitherCatagory (pszId);

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getDitherCatagory: it returns " << pszRet << endl;
#endif
            }

            g_module_close (hModLibrary);
         }
      }
   }

   return pszRet;
}

bool DeviceDither::
ditherCatagoryValid (char *pszId)
{
   char  achLibraryName[512]; // @TBD
   char  achFN[512];          // @TBD
   char *pszLibraryName      = 0;
   bool  fRet;

   fRet = GplDitherInstance::ditherCatagoryValid (pszId);

   if (!fRet)
   {
      pszLibraryName = queryLibrary (achLibraryName, pszId);

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherCatagoryValid: queryLibrary (" << pszId << ") = " << pszLibraryName << endl;
#endif

      if (!g_module_supported ())
      {
         cerr << "DeviceDither::ditherCatagoryValid: This program needs glib's module routines!" << endl;

         return false;
      }

      if (pszLibraryName)
      {
         PFNDITHERCATAGORYVALID  pfnDitherCatagoryValid = 0;
         GModule                *hModLibrary            = 0;

         hModLibrary = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherCatagoryValid: g_module_open (" << pszLibraryName << ") = 0x" << hex << (int)hModLibrary << endl;
#endif

         if (hModLibrary)
         {
#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherCatagoryValid: dlsym (" << convert (achFN, "ditherCatagoryValid", pszLibraryName) << ")";
#endif

            g_module_symbol (hModLibrary,
                             convert (achFN,
                                      "ditherCatagoryValid",
                                      pszLibraryName),
                             (gpointer *)&pfnDitherCatagoryValid);

#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << " = 0x" << hex << (int)pfnDitherCatagoryValid << endl;
#endif

            if (pfnDitherCatagoryValid)
            {
               fRet = pfnDitherCatagoryValid (pszId);

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherCatagoryValid: it returns " << fRet << endl;
#endif
            }

            g_module_close (hModLibrary);
         }
      }
   }

   return fRet;
}

Enumeration * DeviceDither::
getEnumeration ()
{
   class DitherEnumerator : public Enumeration
   {
   private:
      typedef struct _NameEntry {
         struct _NameEntry *pNext;
         char               achName[1];
      } __attribute__ ((aligned (1))) __attribute__ ((packed)) NAMEENTRY, *PNAMEENTRY;

      #define NAMEENTRY_HEADER (sizeof (((PNAMEENTRY)0)->pNext))
      #define NAMEENTRY_BLOCK  4096

   public:
      DitherEnumerator ()
      {
         pCurrentEnum_d = GplDitherInstance::getEnumeration ();
         pNames_d       = (PNAMEENTRY)calloc (1, NAMEENTRY_BLOCK); // fix from manolo@NCC-1701.B.shuttle.de
         pCurrentName_d = pNames_d;
         hLibrary_d     = 0;

         if (pNames_d)
         {
            int        cbNames            = NAMEENTRY_BLOCK;
            int        cbAllocated        = NAMEENTRY_BLOCK;
            char       achLine[512]; // @TBD
            ifstream   ifIn ("/etc/omni");
            char      *pszLine;
            PNAMEENTRY pCurrent           = pNames_d;

            while (0 < ifIn.getline (achLine, sizeof (achLine)))
            {
               pszLine = achLine;

               // Skip whitespace
               while (isspace (*pszLine))
                  pszLine++;

               if ('#' == *pszLine)
               {
                  // Found a comment
                  continue;
               }
               else if (0 != strncmp (pszLine, "dither ", 7))
               {
#ifndef RETAIL
                  if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither: Found non-dither " << pszLine << endl;
#endif
                  continue;
               }

               pszLine += 7;

               // Skip the LHS
               while (*pszLine && !isspace (*pszLine))
                  pszLine++;

               while (!isspace (*pszLine))
                  continue;

               // Skip whitespace
               while (isspace (*pszLine))
                  pszLine++;

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getEnumeration: Testing: " << pszLine << endl;
#endif

               if (!ditherLibraryValid (pszLine))
                  continue;

               if (elementExists (pNames_d, pszLine))
                  continue;

               int iLength = 6                // lib.so
                           + strlen (pszLine)
                           + 1;               // end-of-string

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getEnumeration: Adding: " << pszLine << endl;
#endif

               // Is there room of an element plus an empty end-of-list one?
               if (cbNames > iLength + (int)sizeof (NAMEENTRY))
               {
                  pCurrent->pNext = (PNAMEENTRY)(pCurrent->achName + iLength);
                  sprintf (pCurrent->achName, "lib%s.so", pszLine);
                  cbNames -= iLength + NAMEENTRY_HEADER;

                  pCurrent = pCurrent->pNext;

                  // Terminate the list
                  pCurrent->pNext = 0;
                  pCurrent->achName[0] = '\0';
               }
               else
               {
                  PNAMEENTRY pTemp;
                  PNAMEENTRY pFrom;
                  PNAMEENTRY pTo;

                  cbAllocated += NAMEENTRY_BLOCK;

                  // Allocate a new block
                  pTemp = (PNAMEENTRY)calloc (1, cbAllocated);
                  if (pTemp)
                  {
                     cbNames = cbAllocated;

                     // Copy the old list to the new
                     pFrom = pNames_d;
                     pTo   = pTemp;
                     while (pFrom->achName[0])
                     {
                        iLength = strlen (pFrom->achName) + 1;
                        pTo->pNext = (PNAMEENTRY)(pTo->achName + iLength);
                        strcpy (pTo->achName, pFrom->achName);
                        cbNames -= iLength + NAMEENTRY_HEADER;

                        // Move to the next one
                        pFrom = pFrom->pNext;
                        pTo   = pTo->pNext;
                     }

                     // Add the new element
                     iLength = 6 + strlen (pszLine) + 1;
                     pTo->pNext = (PNAMEENTRY)(pTo->achName + iLength);
                     sprintf (pCurrent->achName, "lib%s.so", pszLine);
                     cbNames -= iLength + NAMEENTRY_HEADER;

                     // Terminate the list
                     pTo->pNext = 0;
                     pTo->achName[0] = 0;

                     // Update the root pointer
                     free (pNames_d);
                     pNames_d = pTemp;
                  }
                  else
                  {
                     // Error!
                     free (pNames_d);
                     pNames_d = 0;
                     break;
                  }
               }
            }
         }
      }

      virtual ~DitherEnumerator ()
      {
         // Clean up
         if (pCurrentEnum_d)
         {
            delete pCurrentEnum_d;
            pCurrentEnum_d = 0;
         }

         if (pNames_d)
         {
            free (pNames_d);
            pNames_d = 0;
         }

         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }
      }

      virtual bool hasMoreElements ()
      {
         if (!pCurrentEnum_d)
            // Nothing to enumerate
            return false;

         if (pCurrentEnum_d->hasMoreElements ())
            // Current enumeration has more elements
            return true;

         // Move to the next one
         // First, clean up
         delete pCurrentEnum_d;
         pCurrentEnum_d = 0;

         if (hLibrary_d)
         {
            g_module_close (hLibrary_d);
            hLibrary_d = 0;
         }

         // Next, loop through our list of libraries
         while (  pCurrentName_d
               && pCurrentName_d->achName[0]
               )
         {
            char               achFN[512];          // @TBD
            PFNGETENUMERATION  pfnGetEnumeration;
            char              *pszLibraryName;

            pszLibraryName = pCurrentName_d->achName;

            // Load it
            hLibrary_d = g_module_open (pszLibraryName, (GModuleFlags)0);

#ifndef RETAIL
            if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getEnumeration: g_module_open (" << pszLibraryName << ") = 0x" << hex << (int)hLibrary_d << endl;
#endif

            if (hLibrary_d)
            {
#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getEnumeration: dlsym (" << convert (achFN, "getEnumeration", pszLibraryName) << ")";
#endif

               // Get the enumeration function
               g_module_symbol (hLibrary_d,
                                convert (achFN,
                                         "getEnumeration",
                                         pszLibraryName),
                                (gpointer *)&pfnGetEnumeration);

#ifndef RETAIL
               if (DebugOutput::shouldOutputDeviceDither ()) cerr << " = 0x" << hex << (int)pfnGetEnumeration << endl;
#endif

               if (pfnGetEnumeration)
               {
                  // Call it
                  pCurrentEnum_d = pfnGetEnumeration ();

#ifndef RETAIL
                  if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::getEnumeration: pCurrentEnum_d = 0x" << hex << (int)pCurrentEnum_d << endl;
#endif
               }
            }

            // Move to the next name in the list
            pCurrentName_d = pCurrentName_d->pNext;

            if (pCurrentEnum_d)
               // Success
               return true;
         }

         return false;
      }

      virtual void * nextElement ()
      {
         void *pvRet = 0;

         if (pCurrentEnum_d)
         {
            pvRet = pCurrentEnum_d->nextElement ();
         }

         return pvRet;
      }

   private:
      bool elementExists (PNAMEENTRY pRoot, char *pszElement)
      {
         PNAMEENTRY pItem = pRoot;

         if (!pItem)
            return false;

         while (pItem->achName[0])
         {
            if (0 == strncmp (pItem->achName + 3,   // Skip the "lib" head
                              pszElement,
                              strlen (pszElement))) // Only test for the name (which doesnt have the ".so" tail)
               return true;

            pItem = pItem->pNext;
         }

         return false;
      }

      Enumeration *pCurrentEnum_d;
      PNAMEENTRY   pNames_d;
      PNAMEENTRY   pCurrentName_d;
      GModule     *hLibrary_d;
   };

   if (!g_module_supported ())
   {
      cerr << "DeviceDither::getEnumeration: This program needs glib's module routines!" << endl;

      return 0;
   }

   return new DitherEnumerator ();
}

// 00000518 T createDitherInstance__3BobPcP6DeviceT1
// 0000336c T ditherRGBtoCMYK__3BobP11_BitmapInfoPUc
// 00003374 T ditherAllPlanesBlank__3Bob
// 00003380 T ditherCPlaneBlank__3Bob
// 000033a4 T ditherKPlaneBlank__3Bob
// 0000338c T ditherMPlaneBlank__3Bob
// 00003398 T ditherYPlaneBlank__3Bob
// 000033b0 T newFrame__3Bob
// 000033b8 T getCPlane__3Bob
// 000033c4 T getMPlane__3Bob
// 000033d0 T getYPlane__3Bob
// 000033dc T getKPlane__3Bob
// 000033e8 T ditherNameValid__3BobPc
// 000033f4 T IDToName__3Bobi
// 00003404 T getDitherCatagory__3BobPc
// 00003414 T ditherCatagoryValid__FPc
// 00003424 T getEnumeration__3Bob

static char *apszRoutines[] = {
   "createDitherInstance__%d%sPcP6DeviceT1",
   "ditherRGBtoCMYK__%d%sP11_BitmapInfoPUc",
   "ditherAllPlanesBlank__%d%s",
   "ditherCPlaneBlank__%d%s",
   "ditherKPlaneBlank__%d%s",
   "ditherMPlaneBlank__%d%s",
   "ditherYPlaneBlank__%d%s",
   "ditherLCPlaneBlank__%d%s",
   "ditherLMPlaneBlank__%d%s",
   "newFrame__%d%s",
   "getCPlane__%d%s",
   "getMPlane__%d%s",
   "getYPlane__%d%s",
   "getKPlane__%d%s",
   "getLCPlane__%d%s",
   "getLMPlane__%d%s",
   "ditherNameValid__%d%sPc",
   "getDitherCatagory__%d%sPc",
   "ditherCatagoryValid__%d%sPc",
   "getEnumeration__%d%s"
};

bool
ditherLibraryValid (char *pszName)
{
   char         achFileName[512]; // @TBD
   GModule     *hDevice;
   bool         fValid           = true;

   if (  !pszName
      || !*pszName
      )
   {
      return false;
   }

   if (!g_module_supported ())
   {
      cerr << "ditherLibraryValid: This program needs glib's module routines!" << endl;

      return false;
   }

   int cbName = strlen (pszName);

   sprintf (achFileName, "lib%s.so", pszName);

   hDevice = g_module_open (achFileName, (GModuleFlags)0);

   if (hDevice)
   {
      gpointer ptr;

      for (int i = 0; i < (int)dimof (apszRoutines); i++)
      {
         sprintf (achFileName, apszRoutines[i], cbName, pszName);

         if (!g_module_symbol (hDevice, achFileName, &ptr))
         {
            fValid = false;
         }

#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::ditherLibraryValid: dlsym (" << achFileName << ") returns 0x" << hex << (int)ptr << endl;
#endif
      }

      g_module_close (hDevice);
   }
   else
   {
      cerr << "DeviceDither::ditherLibraryValid: g_module_error returns " << g_module_error () << endl;

      fValid = false;
   }

   return fValid;
}

char *
queryLibrary (char *pszFillIn, char *pszID)
{
   char       achLine[512]; // @TBD
   ifstream   ifIn ("/etc/omni");
   char      *pszLine;
   char      *pszLHS;
   char      *pszRHS;
   char      *pszSplit;

   *pszFillIn = '\0';

   while (0 < ifIn.getline (achLine, sizeof (achLine)))
   {
      pszLine = achLine;

      // Skip whitespace
      while (isspace (*pszLine))
         pszLine++;

      if ('#' == *pszLine)
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::queryLibrary: Found comment " << pszLine << endl;
#endif
         continue;
      }
      else if (0 != strncmp (pszLine, "dither ", 7))
      {
#ifndef RETAIL
         if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::queryLibrary: Found non-dither " << pszLine << endl;
#endif
         continue;
      }

      pszLine = pszLine + 7;
      pszLHS  = pszLine;

      while (*pszLine && !isspace (*pszLine))
         pszLine++;

      while (!isspace (*pszLine))
         continue;

      pszSplit = pszLine;

      // Skip whitespace
      while (isspace (*pszLine))
         pszLine++;

      pszRHS = pszLine;
      *pszSplit = '\0';

#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::queryLibrary: pszLHS = " << pszLHS << endl;
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::queryLibrary: pszRHS = " << pszRHS << endl;
#endif

      // found a match?
#ifndef RETAIL
      if (DebugOutput::shouldOutputDeviceDither ()) cerr << "DeviceDither::queryLibrary: Testing \"" << pszID << "\" with \"" << pszLHS << "\"" << endl;
#endif
      if (0 != strcmp (pszID, pszLHS))
         continue;

      if (ditherLibraryValid (pszRHS))
      {
         sprintf (pszFillIn, "lib%s.so", pszRHS);

         return pszFillIn;
      }
   }

   return 0;
}

char *
convert (char *pszFillIn, char *pszPFN, char *pszLibraryName)
{
   strcpy (pszFillIn, pszPFN);

   // Loop through the routines that we know about
   for (int i = 0; i < (int)dimof (apszRoutines); i++)
   {
      // Match?
      if (0 == strncmp (pszPFN, apszRoutines[i], strlen (pszPFN)))
      {
         char  achLibraryName[512]; // @TBD

         strcpy (achLibraryName, pszLibraryName);

         pszLibraryName = truncate (achLibraryName);

         // fill out the entry
         sprintf (pszFillIn, apszRoutines[i], strlen (pszLibraryName), pszLibraryName);

         break;
      }
   }

   return pszFillIn;
}
