#include "DriverInfo.hpp"
#include "Device.hpp"
#include "Omni.hpp"

#include "JobAPI.hpp"

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>
#include <vector>

#include <glib.h>
#include <gmodule.h>

/**********************************************************************/

vector<string>
DriverInfo::listOrientations()
{
	vector<string> list;	
	DeviceOrientation *pDO   = pDevice->getCurrentOrientation ();
	Enumeration       *pEnum = pDO->getEnumeration ();
	
	
	list.push_back(DeviceOrientation::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DeviceOrientation *)pEnum->nextElement ();
		list.push_back(DeviceOrientation::IDToName (pDO->getID ()) );
	 	
		// strip the property name here 
		
		delete pDO;
	}
	delete pEnum;

	return list;
}

vector<string>
DriverInfo::listForms()
{
	vector<string> list;	
	DeviceForm *pDO   = pDevice->getCurrentForm ();
	Enumeration       *pEnum = pDO->getEnumeration ();

       	list.push_back( DeviceForm::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DeviceForm *)pEnum->nextElement ();
		list.push_back(DeviceForm::IDToName (pDO->getID ()));
	 	
		// strip the property name here 
		
		delete pDO;
	}
	
	delete pEnum;

	return list;
}

vector<string>
DriverInfo::listTrays()
{
	vector<string> list;	
	DeviceTray *pDO   = pDevice->getCurrentTray ();
	Enumeration       *pEnum = pDO->getEnumeration ();

       	list.push_back(DeviceTray::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DeviceTray *)pEnum->nextElement ();
		list.push_back(DeviceTray::IDToName (pDO->getID ()));
	 	
		// strip the property name here 
		
		delete pDO;
	}
	
	delete pEnum;

	return list;
}


vector<string>
DriverInfo::listMedias()
{
	vector<string> list;	
	DeviceMedia *pDO   = pDevice->getCurrentMedia ();
	Enumeration       *pEnum = pDO->getEnumeration ();

       	list.push_back(DeviceMedia::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DeviceMedia *)pEnum->nextElement ();
		list.push_back(DeviceMedia::IDToName (pDO->getID ()));
	 	
		// strip the property name here 
		
		delete pDO;
	}
	
	delete pEnum;

	return list;
}

vector<string>
DriverInfo::listResolutions()
{
	vector<string> list;	
	DeviceResolution *pDO   = pDevice->getCurrentResolution ();
	Enumeration       *pEnum = pDO->getEnumeration ();

       	list.push_back(DeviceResolution::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DeviceResolution *)pEnum->nextElement ();
		list.push_back(DeviceResolution::IDToName (pDO->getID ()));
	 	
		// strip the property name here 
		
		delete pDO;
	}
	
	delete pEnum;

	return list;
}


vector<string>
DriverInfo::listPrintModes()
{
	vector<string> list;	
	DevicePrintMode *pDO   = pDevice->getCurrentPrintMode ();
	Enumeration       *pEnum = pDO->getEnumeration ();

       	list.push_back(DevicePrintMode::IDToName (pDO->getID ()));  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		pDO = (DevicePrintMode *)pEnum->nextElement ();
		list.push_back(DevicePrintMode::IDToName (pDO->getID ()));
	 	
		// strip the property name here 
		
		delete pDO;
	}
	
	delete pEnum;

	return list;
}

vector<string>
DriverInfo::listDithers()
{
	vector<string> list;	

	Enumeration       *pEnum = DeviceDither::getEnumeration ();

       	list.push_back(pDevice->getCurrentDitherID ());  // default value
	for(int i=1; pEnum->hasMoreElements () ;i++)
	{
		list.push_back((char *)pEnum->nextElement ()) ;
	}

	delete pEnum;

	return list;
}	



void 
DriverInfo::generateDriverInfo()
{
	vector<string> list(25);
	int i;
	int propertyCount = 1;

	
	list = listOrientations ();
	dps.push_back( DriverProperty("orientation","Orientation","ComboBox",list,list[0]));

	list = listForms ();
	dps.push_back(DriverProperty("form","Form","ComboBox",list,list[0]));

	list = listTrays ();
	dps.push_back( DriverProperty("tray","Tray","ComboBox",list,list[0]));
	
	list = listMedias ();
	dps.push_back( DriverProperty("media","Media","ComboBox",list,list[0]));
	
	list = listResolutions ();
	dps.push_back( DriverProperty("resolution","Resolution","ComboBox",list,list[0]));

	list = listDithers ();
	dps.push_back(DriverProperty("dither","Dither","ComboBox",list,list[0]));

	list = listPrintModes ();
	dps.push_back( DriverProperty("printMode","Print Mode","ComboBox",list,list[0]));

	
	driver = Driver("First Driver", dps);

	return;
}

int
DriverInfo::openDevice (char *pszFullDeviceName)
{

  hmodDevice = g_module_open (pszFullDeviceName, (GModuleFlags)0);

   if (!hmodDevice)
   {
      cerr << "g_module_error returns " << g_module_error () << endl;
      return 2;
   }
  	
   g_module_symbol (hmodDevice, "newDevice__Fb", (gpointer *)&pfnNewDevice);

   if (!pfnNewDevice)
      return 3;

   pDevice = pfnNewDevice (true);
   
   return 0;

}

/***********************************************************/

vector<string> DriverInfo::
getDeviceList()
{
	
	vector<string> deviceList;
	Enumeration *pEnum = Omni::listDevices ();
	
	cerr<<"\n Enumerating the devices. Please wait ... (It may take few seconds)"<<endl;
	while (pEnum->hasMoreElements ())
	{
		char *pszLibName = (char *)pEnum->nextElement ();
		 deviceList.push_back(pszLibName);
	}	 
	
	delete pEnum;

	cerr<<endl<<"************************************************"<<endl;
	for(int iz=0;iz<deviceList.size();iz++)
	{
		int point = deviceList[iz].find_last_of('/');
		deviceList[iz] = deviceList[iz].substr(point+4);
		deviceList[iz] = deviceList[iz].erase(deviceList[iz].size()-3);
	}
	
	return deviceList;
}
	
