#include <vector>
#include <gtk--/main.h>

#include "JobPropertyDialogController.hpp"
#include "JobPropertyDialogView.hpp"
#include "WidgetManager.hpp"

using namespace Gtk;
using namespace JobDialog;

vector<string> JobPropertyDialogController::getSelectedJobProperties()
{
	
	setupDialog() ; 
	Kit::run();   // After setting up the dialog this starts an event loop and waits for the vents to happen.
		
	vector<string> propertyValueList;
	vector<PropertyWidgetPair>::iterator itr;
	
	for(itr=propertyWidgetList.begin(); itr!=propertyWidgetList.end(); itr++)
	{
		DriverProperty* dp = itr->driverProperty;
		propertyValueList.push_back(dp->getName()+"="+dp->getDefaultValue() );
	}
	
	return propertyValueList;
}

vector<string> JobPropertyDialogController::getDefaultJobProperties()
{
	
	vector<string> propertyValueList;
	
	vector<string> propertyNames = driver.getAvailablePropertyNames();
	vector<string>::iterator itr;
		
	for(itr=propertyNames.begin(); itr !=propertyNames.end(); itr++)
	{
		DriverProperty dp = driver.getProperty(*itr);
		propertyValueList.push_back(dp.getName()+"="+dp.getDefaultValue() );
	}
	
	return propertyValueList;
}

void JobPropertyDialogController::setupDialog()
{

	vector<string> propertyNames = driver.getAvailablePropertyNames();
	view = new JobPropertyDialogView (this, (int)propertyNames.size());
	
	// Iterate thru the vector , get the DriverProperty, get the widget made for that with the help of 
	// WidgetManager. Then give this populated widget to View and ask it to add that widget with a label.
	// Finally maintain the widget in a list along with DriverProperty object.
	
	vector<string>::iterator itr;
	PropertyWidgetPair pair;	
	
	for(itr=propertyNames.begin(); itr !=propertyNames.end(); itr++)
	{
		DriverProperty dp = driver.getProperty(*itr);
		Widget *widget  = WidgetManager::instance()->createWidget(dp.getWidgetName(), &dp);
		
		view->addWidget(widget, dp.getPropertyNameForDisplay());
		
		pair.driverProperty = new DriverProperty(dp) ;
		pair.widget = widget;
		pair.defaultValue =dp.getDefaultValue();
		
		propertyWidgetList.push_back(pair);
	}	
	view->showDialog();

	return;
}

void JobPropertyDialogController::onPrintButtonClicked()
{
	vector<PropertyWidgetPair>::iterator itr;
	for(itr=propertyWidgetList.begin(); itr!=propertyWidgetList.end(); itr++)
	{
		DriverProperty *dp = itr->driverProperty ; 
		string newDefaultValue= WidgetManager::instance()->getSelectedValue(itr->widget , dp->getWidgetName() );
		dp->setDefaultValue(newDefaultValue);
		
	}	
	
	Kit::quit();	// stop the event loop and close the Dialog. 
	return;
	
}

void JobPropertyDialogController::onSaveButtonClicked()
{
	vector<PropertyWidgetPair>::iterator itr;
	for(itr=propertyWidgetList.begin(); itr!=propertyWidgetList.end(); itr++)
	{
		DriverProperty *dp = itr->driverProperty ; 
		string newDefaultValue= WidgetManager::instance()->getSelectedValue(itr->widget , dp->getWidgetName() );
		dp->setDefaultValue(newDefaultValue);
		driver.setProperty(dp->getName(),*dp);
	}
	
	return;   // Don't stop the event loop. This is eqvt. to pressing Apply button. 
}

void JobPropertyDialogController::onSaveNprintButtonClicked()
{
	// Eqvt. to Save click as SaveHandler includes the work PrintHandler does. 

	onSaveButtonClicked();
	Kit::quit();					 		// stop the event loop and close the Dialog. 
	 
	return;
}

void JobPropertyDialogController::onCancelButtonClicked()  
{
	// Do nothing right now ...

	 Kit::quit();			// stop the event loop and close the Dialog. 
	 
	return;
}

JobPropertyDialogController::~JobPropertyDialogController()
{
	delete view;
	
	vector<PropertyWidgetPair>::iterator itr;
	for(itr=propertyWidgetList.begin(); itr!=propertyWidgetList.end(); itr++)
	{
		delete itr->driverProperty ;
	}
}
