#include "JobPropertyDialogView.hpp"
#include "JobPropertyDialogController.hpp"

using namespace Gtk;
using namespace JobDialog;

JobPropertyDialogView::JobPropertyDialogView(JobPropertyDialogController *ctr, int pc) : controller(ctr), totalPropertyCount(pc)
{
	addedPropertyCount = 0;
	vbox = manage ( new VBox() );
	
	scrWindow = manage(new ScrolledWindow() );	
	table = manage(new Table(totalPropertyCount,2) );
	frame =manage(new Frame());
	hbox = manage(new HButtonBox() );
	
	saveButton = 			manage(new Button("Save") );
	saveNprintButton = 	manage(new Button("Save&Print") );
	printButton = 			manage(new Button("Print") );
	cancelButton = 		manage(new Button("Cancel") );
	
	createSkeleton();
	setupConnections();
}

bool JobPropertyDialogView::addWidget(Widget* widget, string displayName)
{
	if(addedPropertyCount >= totalPropertyCount)
		return false;
		
	/******************************************************************************************/	
	/*   void Table::attach  (Widget& child, guint left_attach, guint right_attach, guint top_attach,            */
        /*                                   guint bottom_attach, gint xoptions=(GTK_FILL|GTK_EXPAND), gint           */
        /*                           yoptions=(GTK_FILL|GTK_EXPAND), guint xpadding=0, guint ypadding=0);	  */
        /*****************************************************************************************/ 
	
	Label *displayNameLabel  = manage(new Label(displayName) );

	table->attach(*displayNameLabel, 0,1, addedPropertyCount, addedPropertyCount+1, 0,0,0,0);
	table->attach(*widget, 1,2, addedPropertyCount, addedPropertyCount+1, 0,0,0,0);
	
	addedPropertyCount++;
	
	return true;
}

void JobPropertyDialogView::showDialog()
{
	show_all();
	
	return;
}

void JobPropertyDialogView::createSkeleton()
{
	add(*vbox);
	
	vbox->pack_start(*scrWindow);
	vbox->pack_start(*frame,false,false);    
		
	scrWindow->set_policy(GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	
	scrWindow->add_with_viewport(*table);
	
	table->set_row_spacings(15);
	table->set_col_spacings(15);
	table->set_border_width(20);

	frame->add(*hbox);
						
	hbox->set_border_width(10);
	hbox->set_layout(GTK_BUTTONBOX_END);
	hbox->set_spacing(20);
	hbox->set_child_size(60,25);

	hbox->pack_start(*saveButton, false);
	hbox->pack_start(*saveNprintButton, false);	
	hbox->pack_start(*printButton, false);	
	hbox->pack_start(*cancelButton, false);	
		
	set_usize(450,400);	
	set_modal(true);
	
	return;	
}

void JobPropertyDialogView::setupConnections()
{
	using SigC::slot;
		
	saveNprintButton->clicked.connect(slot(controller, &JobPropertyDialogController::onSaveNprintButtonClicked) );
	saveButton->clicked.connect(slot(controller, &JobPropertyDialogController::onSaveButtonClicked) );
	printButton->clicked.connect(slot(controller, &JobPropertyDialogController::onPrintButtonClicked) );
	cancelButton->clicked.connect(slot(controller, &JobPropertyDialogController::onCancelButtonClicked) );
	
	return;
}
