#include "WidgetManager.hpp"

using namespace Gtk;
using namespace JobDialog;

WidgetManager* WidgetManager::uniqueInstance = 0;   // 	static member initialization.

WidgetManager* WidgetManager::instance()
{
	if(uniqueInstance == 0) 
		uniqueInstance=new WidgetManager; 

	return uniqueInstance;
}

Widget* WidgetManager::createWidget(string widgetName, DriverProperty *driverProperty)
{
	Widget* widget;
	
	if("TextBox"==widgetName)
		widget = manage(new Entry() );
			
	else if("ComboBox" == widgetName)
		widget = manage(new Combo() );
	
	else
		return 0;	 // Probably its better to throw an Exception here 'UnKnownWidgetException'
		
	WidgetHandler * widgetHandler = getWidgetHandler(widgetName);
	widgetHandler->populateValues(widget, driverProperty);
	
	return widget;	
}

string WidgetManager::getSelectedValue(Widget* widget, string widgetName)
{
	WidgetHandler* widgetHandler = getWidgetHandler(widgetName);

	return widgetHandler->getSelectedValue(widget);	
}

WidgetHandler* WidgetManager::getWidgetHandler(string widgetName)
{
	WidgetHandler *widgetHandler;
	
	if("TextBox"==widgetName)
		widgetHandler = TextBoxHandler::instance();
	
	else if("ComboBox" == widgetName)
		widgetHandler = ComboBoxHandler::instance();	
	
	else
		widgetHandler = 0;  // Probably its better to throw an Exception here 'UnKnownWidgetException'
		
	return widgetHandler;
}
