/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2001
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "Okidata_PCL_Instance.hpp"
#include "defines.hpp"
#include <GplDitherInstance.hpp>

#include <iostream>
#include <strstream>
#include <stdio.h>

Okidata_PCL_Instance::
Okidata_PCL_Instance (Device *pDevice)
   : DeviceInstance (pDevice)
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::Okidata_PCL_Instance ()" << endl;

   ptlPrintHead_d.x     = 0;
   ptlPrintHead_d.y     = 0;
   iUOM_d               = 0;
   iXScalingFactor_d    = 1;
   iYScalingFactor_d    = 1;
   iYUOMScalingFactor_d = 0;
   iVerticalOffset_d    = 0;
   fHaveInitialized_d   = false;
   fHaveSetupPrinter_d  = false;
   iHardwareScaling_d   = 0;
}

Okidata_PCL_Instance::
~Okidata_PCL_Instance ()
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::~Okidata_PCL_Instance ()" << endl;
}

void Okidata_PCL_Instance::
initializeInstance ()
{
   if (fHaveInitialized_d)
      return;

   fHaveInitialized_d = true;

#ifdef DEBUG
if (hasDeviceOption (DO1_AUTO_ROTATE)) cerr << "DO1_AUTO_ROTATE" << endl;
if (hasDeviceOption (DO1_SCALABLE_FONT)) cerr << "DO1_SCALABLE_FONT" << endl;
if (hasDeviceOption (DO1_SUPPORT_DUPLEX)) cerr << "DO1_SUPPORT_DUPLEX" << endl;
if (hasDeviceOption (DO1_SUPPORT_RULES)) cerr << "DO1_SUPPORT_RULES" << endl;
if (hasDeviceOption (DO1_SUPPORT_DOT_COORDS)) cerr << "DO1_SUPPORT_DOT_COORDS" << endl;
if (hasDeviceOption (DO1_SUPPORT_PJL)) cerr << "DO1_SUPPORT_PJL" << endl;
if (hasDeviceOption (DO1_SUPPORT_BACKBIN)) cerr << "DO1_SUPPORT_BACKBIN" << endl;
if (hasDeviceOption (DO1_SUPPORT_JOBOFFSET)) cerr << "DO1_SUPPORT_JOBOFFSET" << endl;
if (hasDeviceOption (DO1_SUPPORT_PATTERNS)) cerr << "DO1_SUPPORT_PATTERNS" << endl;
if (hasDeviceOption (DO1_SUPPORT_PCL5)) cerr << "DO1_SUPPORT_PCL5" << endl;
if (hasDeviceOption (DO1_MIXED_TRAY)) cerr << "DO1_MIXED_TRAY" << endl;
if (hasDeviceOption (DO1_SUPPORT_COLOR)) cerr << "DO1_SUPPORT_COLOR" << endl;
if (hasDeviceOption (DO1_SUPPORT_PS)) cerr << "DO1_SUPPORT_PS" << endl;
if (hasDeviceOption (DO1_SUPPORT_PAGEPROT_ESC)) cerr << "DO1_SUPPORT_PAGEPROT_ESC" << endl;
if (hasDeviceOption (DO1_SUPPORT_ES)) cerr << "DO1_SUPPORT_ES" << endl;
if (hasDeviceOption (DO1_SUPPORT_HPGL2)) cerr << "DO1_SUPPORT_HPGL2" << endl;
if (hasDeviceOption (DO1_SUPPORT_HPGL2_MC)) cerr << "DO1_SUPPORT_HPGL2_MC" << endl;
if (hasDeviceOption (DO1_PAGEPROT_AUTO)) cerr << "DO1_PAGEPROT_AUTO" << endl;
if (hasDeviceOption (DO1_SUPPORT_COMPRESS5)) cerr << "DO1_SUPPORT_COMPRESS5" << endl;
if (hasDeviceOption (DO1_SUPPORT_COLORSMART)) cerr << "DO1_SUPPORT_COLORSMART" << endl;
if (hasDeviceOption (DO1_SUPPORT_HPGL_DEFAULT)) cerr << "DO1_SUPPORT_HPGL_DEFAULT" << endl;
if (hasDeviceOption (DO1_SUPPORT_OUTPUTOPTION)) cerr << "DO1_SUPPORT_OUTPUTOPTION" << endl;
if (hasDeviceOption (DO1_PRINTER_DISPLAY)) cerr << "DO1_PRINTER_DISPLAY" << endl;
if (hasDeviceOption (DO1_SUPPORT_COLOR_RET)) cerr << "DO1_SUPPORT_COLOR_RET" << endl;
if (hasDeviceOption (DO1_SUPPORT_HGRAYSCALE)) cerr << "DO1_SUPPORT_HGRAYSCALE" << endl;
if (hasDeviceOption (DO1_SUPPORT_WATERMARK)) cerr << "DO1_SUPPORT_WATERMARK" << endl;
if (hasDeviceOption (DO1_SUPPORT_HPGL2MACRO)) cerr << "DO1_SUPPORT_HPGL2MACRO" << endl;
if (hasDeviceOption (DO1_SUPPORT_PAN_SUPER)) cerr << "DO1_SUPPORT_PAN_SUPER" << endl;
if (hasDeviceOption (DO1_SUPPORT_CONFIG)) cerr << "DO1_SUPPORT_CONFIG" << endl;
if (hasDeviceOption (DO1_SUPPORT_HPGL2LOW)) cerr << "DO1_SUPPORT_HPGL2LOW" << endl;
if (hasDeviceOption (DO1_SUPPORT_BIDI)) cerr << "DO1_SUPPORT_BIDI" << endl;
if (hasDeviceOption (DO1_SUPPORT_STAPLER)) cerr << "DO1_SUPPORT_STAPLER" << endl;

if (hasDeviceOption2 (DO2_SUPPORT_PCLXL)) cerr << "DO2_SUPPORT_PCLXL" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_AUTOSELECT)) cerr << "DO2_SUPPORT_AUTOSELECT" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCLXL_DEFAULT)) cerr << "DO2_SUPPORT_PCLXL_DEFAULT" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCLXL_1200)) cerr << "DO2_SUPPORT_PCLXL_1200" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_LARGECLRFMT)) cerr << "DO2_SUPPORT_LARGECLRFMT" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_2STAPLES)) cerr << "DO2_SUPPORT_2STAPLES" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCLXL_600)) cerr << "DO2_SUPPORT_PCLXL_600" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_NEWCLR)) cerr << "DO2_SUPPORT_NEWCLR" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PJLOUTPUTBIN)) cerr << "DO2_SUPPORT_PJLOUTPUTBIN" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_COLLATOR)) cerr << "DO2_SUPPORT_COLLATOR" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_MET)) cerr << "DO2_SUPPORT_MET" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_MOPIER)) cerr << "DO2_SUPPORT_MOPIER" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_MOPIER_DEFAULT)) cerr << "DO2_SUPPORT_MOPIER_DEFAULT" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_EUROCHAR)) cerr << "DO2_SUPPORT_EUROCHAR" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_JOB_RETENTION)) cerr << "DO2_SUPPORT_JOB_RETENTION" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCLXL_1200R)) cerr << "DO2_SUPPORT_PCLXL_1200R" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_FAX)) cerr << "DO2_SUPPORT_FAX" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_SORTER)) cerr << "DO2_SUPPORT_SORTER" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_COLOR_OEM)) cerr << "DO2_SUPPORT_COLOR_OEM" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_COLLATE)) cerr << "DO2_SUPPORT_COLLATE" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_QUIETRESET)) cerr << "DO2_SUPPORT_QUIETRESET" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_EDGEPRINTING)) cerr << "DO2_SUPPORT_EDGEPRINTING" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_OPTIMIZEPQ)) cerr << "DO2_SUPPORT_OPTIMIZEPQ" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_COLORSMART2)) cerr << "DO2_SUPPORT_COLORSMART2" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_MANUALDUPLEX)) cerr << "DO2_SUPPORT_MANUALDUPLEX" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_NOFNTMGR)) cerr << "DO2_SUPPORT_NOFNTMGR" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PRINTERSELECTDEF)) cerr << "DO2_SUPPORT_PRINTERSELECTDEF" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCL6_REV20)) cerr << "DO2_SUPPORT_PCL6_REV20" << endl;
if (hasDeviceOption2 (DO2_SUPPORT_PCL6_INT)) cerr << "DO2_SUPPORT_PCL6_INT" << endl;
#endif

   int iValidUnits[] = {
       96,  100,  120,  144,  150,  160,  180,  200,
      225,  240,  288,  300,  360,  400,  450,  480,
      600,  720,  800,  900, 1200, 1440, 1800, 2400,
      3600, 7200
   };
   DeviceResolution *pDR   = getCurrentResolution ();
   int               iYRes = pDR->getExternalYRes ();
   int               iLow  = 0;
   int               iMid  = (int)dimof (iValidUnits) / 2;
   int               iHigh = (int)dimof (iValidUnits) - 1;

   // Search in the list of supported units
   while (iLow <= iHigh)
   {
      if (iYRes == iValidUnits[iMid])
      {
         // Found!
         break;
      }
      else if (iYRes < iValidUnits[iMid])
      {
         iHigh = iMid - 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
      else // (iYRes > iValidUnits[iMid])
      {
         iLow  = iMid + 1;
         iMid  = iLow + (iHigh - iLow) / 2;
      }
   }

   iYUOMScalingFactor_d = 1;

   if (iLow > iHigh)
   {
      // Not found in the list, try for a multiple
      for (iMid = 0; iMid < (int)dimof (iValidUnits); iMid++)
      {
         if (0 == iValidUnits[iMid] % iYRes)
         {
            iYUOMScalingFactor_d = iValidUnits[iMid] / iYRes;
            break;
         }
      }

      if (iMid == (int)dimof (iValidUnits))
      {
         // Still not found.  Default to 300
         if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance: Error unable to handle " << iYRes << "!" << endl;
         iMid = 11; // 300
      }
   }

   HardCopyCap *pHCC     = getCurrentForm ()->getHardCopyCap ();
   int          iTopClip = (int)(((float)pHCC->getTopClip () / 25400.0) * pDR->getExternalYRes () + 0.5);

   iUOM_d            = iValidUnits[iMid];
///iVerticalOffset_d = iUOM_d / 6; // is 50 @ 300 uom
   iVerticalOffset_d = iTopClip;

   if (pDR->getExternalXRes () > pDR->getXRes ())
   {
      iXScalingFactor_d = pDR->getExternalXRes () / pDR->getXRes ();
   }
   if (pDR->getExternalYRes () > pDR->getYRes ())
   {
      iYScalingFactor_d = pDR->getExternalYRes () / pDR->getYRes ();
   }

   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance () iXScalingFactor_d    = " << iXScalingFactor_d << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance () iYScalingFactor_d    = " << iYScalingFactor_d << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance () iUOMYScalingFactor_d = " << iYUOMScalingFactor_d << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance () iUOM_d               = " << iUOM_d << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::initializeInstance () iVerticalOffset_d    = " << iVerticalOffset_d << endl;
}

static char *apszDeviceJobPropertyKeys[] = {
   "HardwareScaling"
};
#define JOBPROP_HWSCALING apszDeviceJobPropertyKeys[0]

Enumeration * Okidata_PCL_Instance::
listDeviceJobPropertyKeys ()
{
   class DeviceJobPropertyKeyEnumerator : public Enumeration
   {
   public:
      DeviceJobPropertyKeyEnumerator (char **apszKeys, int iDimOfKeys)
      {
         iWhichKey_d = iDimOfKeys;
         apszKeys_d  = apszKeys;
      }

      virtual bool hasMoreElements ()
      {
         return iWhichKey_d > 0;
      }

      virtual void * nextElement ()
      {
         if (0 < iWhichKey_d)
         {
            iWhichKey_d--;
            return apszKeys_d[iWhichKey_d];
         }

         return 0;
      }

   private:
      int    iWhichKey_d;
      char **apszKeys_d;
   };

   return new DeviceJobPropertyKeyEnumerator (apszDeviceJobPropertyKeys, dimof (apszDeviceJobPropertyKeys));
}

bool Okidata_PCL_Instance::
handleDeviceJobProperty (char *pszKey,
                         char *pszValue)
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::handleDeviceJobProperty ("
                          << pszKey << ", "
                          << pszValue << ")"
                          << endl;

   if (0 == strcmp (JOBPROP_HWSCALING, pszKey))
   {
      int iValue = atoi (pszValue);

      if (iValue > 0)
      {
         DeviceResolution *pDR   = pDevice_d->getCurrentResolution ();
         int               iXRes = pDR->getXRes ();
         int               iYRes = pDR->getYRes ();

         if (  (iXRes % iValue) == 0
            && (iYRes % iValue) == 0
            )
         {
            iHardwareScaling_d = iValue;

            pDR->setInternalXRes (iXRes / iValue);
            pDR->setInternalYRes (iYRes / iValue);

            return true;
         }
      }
      else if (iValue == 0)
      {
         DeviceResolution *pDR = pDevice_d->getCurrentResolution ();

         iHardwareScaling_d = iValue;

         pDR->setInternalXRes (pDR->getXRes ());
         pDR->setInternalYRes (pDR->getYRes ());

         return true;
      }
   }

   return false;
}

string * Okidata_PCL_Instance::
getJobPropertyType (char *pszKey)
{
   if (0 == strcmp (JOBPROP_HWSCALING, pszKey))
   {
      ostrstream oss;

      oss << "integer "      // type
          << 1               // default
          << " "
          << 1               // minimum
          << ends;

      string *pRet = new string (oss.str ());

      return pRet;
   }

   return 0;
}

string * Okidata_PCL_Instance::
getJobProperty (char *pszKey)
{
   if (0 == strcmp (JOBPROP_HWSCALING, pszKey))
   {
      ostrstream oss;

      oss << iHardwareScaling_d << ends;

      string *pRet = new string (oss.str ());

      return pRet;
   }

   return 0;
}

void Okidata_PCL_Instance::
setupPrinter ()
{
   if (fHaveSetupPrinter_d)
      return;

   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::setupPrinter ()" << endl;

   fHaveSetupPrinter_d = true;

   DeviceResolution *pRes      = getCurrentResolution ();
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   pCmd = pCommands->getCommandData ("cmdSetUnitsOfMeasure");
   if (pCmd)
   {
      pDevice_d->sendPrintfToDevice (pCmd, iUOM_d);
   }
   else
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << dec << "Error: There is no cmdSetUnitsOfMeasure defined for this device!" << endl;
   }

   // Select the form
   pDevice_d->sendBinaryDataToDevice (getCurrentForm ());

   // Select the tray
   pDevice_d->sendBinaryDataToDevice (getCurrentTray ());

   // Select the resolution
   pDevice_d->sendPrintfToDevice (pRes->getData (), pRes->getExternalXRes ());

   pCmd = pCommands->getCommandData ("cmdSetTopMargin");
   if (pCmd)
   {
      pDevice_d->sendBinaryDataToDevice (pCmd);
   }
   else
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << dec << "Error: There is no cmdSetTopMargin defined for this device!" << endl;
   }

   pCmd = pCommands->getCommandData ("cmdSetXYPos");
   if (pCmd)
   {
      pDevice_d->sendPrintfToDevice (pCmd,
                                     0,
                                     iVerticalOffset_d * iYUOMScalingFactor_d);
   }
   else
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << dec << "Error: There is no cmdSetXYPos defined for this device!" << endl;
   }

   switch (getCurrentPrintMode ()->getColorTech ())
   {
   case DevicePrintMode::COLOR_TECH_RGB:
   {
      pCmd = pCommands->getCommandData ("cmdConfigureImageData24BPP");
      if (pCmd)
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);

         BinaryData  *pCmd2   = 0;
         PBYTE        pbGamma = 0;
         DeviceGamma *pGamma  = pDevice_d->getCurrentGamma ();

         pCmd    = pCommands->getCommandData ("cmdSetColorLookupTable");
         pCmd2   = pCommands->getCommandData ("cmdSetColorLookupTableDataParm");
         pbGamma = (PBYTE)malloc (256);

         if (DebugOutput::shouldOutputInstance ()) cerr << "pCmd = " << *pCmd << endl;
         if (DebugOutput::shouldOutputInstance ()) cerr << "pCmd2 = " << *pCmd2 << endl;
         if (DebugOutput::shouldOutputInstance ()) cerr << "pbGamma = 0x" << hex << (int)pbGamma << dec << endl;
         if (DebugOutput::shouldOutputInstance ()) cerr << "pGamma = " << *pGamma << endl;

         if (  pCmd
            && pCmd2
            && pbGamma
            )
         {
            pDevice_d->sendPrintfToDevice (pCmd, 770); // length of following

            pDevice_d->sendPrintfToDevice (pCmd2, 0); // RGB
            pDevice_d->sendPrintfToDevice (pCmd2, 0); // reserved
            for (int i = 0; i < 3; i++)
            {
               switch (i)
               {
               case 0:
               {
                  // Blue gamma values
                  GplGenerateGammaCurve (pGamma->getYGamma (),
                                         pGamma->getYBias (),
                                         pbGamma);
                  break;
               }

               case 1:
               {
                  // Green gamma values
                  GplGenerateGammaCurve (pGamma->getMGamma (),
                                         pGamma->getMBias (),
                                         pbGamma);
                  break;
               }

               case 2:
               {
                  // Red gamma values
                  GplGenerateGammaCurve (pGamma->getCGamma (),
                                         pGamma->getCBias (),
                                         pbGamma);
                  break;
               }
               }

               for (int j = 0; j < 256; j++)
               {
                  if (DebugOutput::shouldOutputInstance ()) cerr << "i = " << i << ", j = " << j << ", data = " << (int)pbGamma[j] << endl;
                  pDevice_d->sendPrintfToDevice (pCmd2, pbGamma[255 - j]);
               }
            }
         }

         if (pbGamma)
            free (pbGamma);
      }
      else
      {
         if (DebugOutput::shouldOutputInstance ()) cerr << dec << "Error: There is no cmdConfigureImageData defined for this device!" << endl;
      }
      break;
   }
   case DevicePrintMode::COLOR_TECH_K:
   {
      pCmd = pCommands->getCommandData ("cmdConfigureImageData1BPP");
      if (pCmd)
      {
         pDevice_d->sendBinaryDataToDevice (pCmd);
      }
      break;
   }
   }
}

bool Okidata_PCL_Instance::
beginJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::beginJob ()" << endl;

   pCmd = pCommands->getCommandData ("cmdEnterLanguage");
   if (  hasDeviceOption (DO1_SUPPORT_PJL)
      && pCmd
      )
   {
      pDevice_d->sendBinaryDataToDevice (pCmd);
   }
   else
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << "Error: There is no cmdEnterLanguage defined for this device!" << endl;
   }

   pCmd = pCommands->getCommandData ("cmdInit");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdInit = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool Okidata_PCL_Instance::
newFrame ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::newFrame ()" << endl;

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   ditherNewFrame ();

   ptlPrintHead_d.x = 0;
   ptlPrintHead_d.y = 0;

   return true;
}

bool Okidata_PCL_Instance::
newFrame (bool fJobPropertiesChanged)
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::newFrame (with props)" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The job properties are as follows:" << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iOrientation = " << getCurrentOrientation() << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "iDitherID = " << getCurrentDitherID () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The form is = " << getCurrentForm () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The tray is = " << getCurrentTray () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The media is = " << getCurrentMedia () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The resolution is = " << getCurrentResolution () << endl;
   if (DebugOutput::shouldOutputInstance ()) cerr << "The print mode is = " << getCurrentPrintMode () << endl;

   // @TBD - reinitialize with new job properties

   // Call common code
   return newFrame ();
}

bool Okidata_PCL_Instance::
endJob ()
{
   DeviceCommand    *pCommands = getCommands ();
   BinaryData       *pCmd      = 0;

   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::endJob ()" << endl;

   pCmd = pCommands->getCommandData ("cmdPageEject");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdPageEject = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   pCmd = pCommands->getCommandData ("cmdTerm");
   if (pCmd)
   {
      if (DebugOutput::shouldOutputInstance ()) cerr << "cmdTerm = " << *pCmd << endl;

      pDevice_d->sendBinaryDataToDevice (pCmd);
   }

   return true;
}

bool Okidata_PCL_Instance::
abortJob ()
{
   if (DebugOutput::shouldOutputInstance ()) cerr << "Okidata_PCL_Instance::endJob ()" << endl;

   return true;
}

string Okidata_PCL_Instance::
toString (ostrstream& oss)
{
   oss << hex
       << "{ "
       << DeviceInstance::toString (oss) // Add parent's output as well
       << dec
       << " }" << ends;

   return oss.str ();
}

/* Provide a way to print out class data
*/
ostream&
operator<< (ostream& os, const Okidata_PCL_Instance& const_self)
{
   Okidata_PCL_Instance& self = const_cast<Okidata_PCL_Instance&>(const_self);
   ostrstream          oss;

   os << self.toString (oss);

   return os;
}
