/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _PrintDevice
#define _PrintDevice

#include "Device.hpp"

#include <glib.h>
#include <gmodule.h>

class PrintDevice : public Device
{
protected:
                              PrintDevice               (char              *pszDriverName,
                                                         char              *pszDeviceName,
                                                         char              *pszShortName,
                                                         char              *pszJobProperties);

public:
   virtual                   ~PrintDevice               ();

   void                       initialize                ();

   char                      *getVersion                ();

   char                      *getDriverName             ();
   char                      *getDeviceName             ();
   char                      *getShortName              ();

   DeviceInstance            *getDeviceInstance         ();

   DeviceOrientation         *getCurrentOrientation     ();
   char                      *getCurrentDitherID        ();
   DeviceDither              *getCurrentDither          ();
   DeviceForm                *getCurrentForm            ();
   DeviceTray                *getCurrentTray            ();
   DeviceMedia               *getCurrentMedia           ();
   DeviceResolution          *getCurrentResolution      ();
   DevicePrintMode           *getCurrentPrintMode       ();
   DeviceCommand             *getCommands               ();
   DeviceData                *getDeviceData             ();
   virtual DeviceGamma       *getCurrentGamma           () = 0;

   int                        getPDLLevel               ();
   int                        getPDLSubLevel            ();
   int                        getPDLMajorRevisionLevel  ();
   int                        getPDLMinorRevisionLevel  ();

   bool                       hasCapability             (long               lMask);
   bool                       hasRasterCapability       (long               lMask);
   bool                       hasDeviceOption           (long               lMask);
   bool                       hasDeviceOption2          (long               lMask);

   int                        getScanlineMultiple       ();

   bool                       beginJob                  ();
   bool                       beginJob                  (char              *pszJobProperties);
   bool                       newFrame                  ();
   bool                       newFrame                  (char              *pszJobProperties);
   bool                       endJob                    ();
   bool                       abortJob                  ();

   bool                       rasterize                 (PBYTE              pbBits,
                                                         PBITMAPINFO2       pbmi,
                                                         PRECTL             prectlPageLocation,
                                                         BITBLT_TYPE        eType);

   void                       setOutputStream           (ostream           *osNew);
   void                       setOutputStream           (int                iFileNo);
   void                       setOutputFunction         (PFNOUTPUTFUNCTION  pfnOutputFunction,
                                                         void              *pMagicCookie);

   bool                       sendBinaryDataToDevice    (BinaryData        *pData);
   bool                       sendBinaryDataToDevice    (DeviceForm        *pForm);
   bool                       sendBinaryDataToDevice    (DeviceTray        *pTray);
   bool                       sendBinaryDataToDevice    (DeviceMedia       *pMedia);
   bool                       sendBinaryDataToDevice    (DeviceResolution  *pResolution);
   bool                       sendBinaryDataToDevice    (PBYTE              pbData,
                                                         int                iLength);

   bool                       sendPrintfToDevice        (BinaryData        *pData,
                                                                            ...);
   bool                       sendVPrintfToDevice       (BinaryData        *pData,
                                                         va_list            list);

   string                    *queryCurrentJobProperties ();
   Enumeration               *listJobPropertyKeys       ();
   Enumeration               *listDeviceJobPropertyKeys ();
   string                    *getJobPropertyType        (char              *pszKey);
   string                    *getJobProperty            (char              *pszKey);

   DeviceBlitter             *getDeviceBlitter          ();

   void                       loadLibrary               (char              *pszLibraryName);
   void                      *dlsym                     (char              *pszLibraryName,
                                                         char              *pszSymbol);

   virtual string             toString                  (ostrstream& oss);
   friend ostream&            operator<<                (ostream& os, const PrintDevice& self);

protected:
   void                       setDeviceInstance         (DeviceInstance   *pInstance);
   void                       setDeviceBlitter          (DeviceBlitter    *pBlitter);
   void                       setPDL                    (PDL              *pPDL);

   void                       setCapabilities           (long              lCapabilities);
   void                       setRasterCapabilities     (long              lCapabilities);
   void                       setDeviceOptions          (long              lOptions);
   void                       setDeviceOptions2         (long              lOptions);

   virtual DeviceOrientation *getDefaultOrientation     () = 0;
   virtual char              *getDefaultDitherID        () = 0;
   virtual DeviceForm        *getDefaultForm            () = 0;
   virtual DeviceTray        *getDefaultTray            () = 0;
   virtual DeviceMedia       *getDefaultMedia           () = 0;
   virtual DeviceResolution  *getDefaultResolution      () = 0;
   virtual DeviceCommand     *getDefaultCommands        () = 0;
   virtual DeviceData        *getDefaultData            () = 0;
   virtual DevicePrintMode   *getDefaultPrintMode       () = 0;

private:
   void                       initializeJobProperties   ();
   void                       cleanupProperties         ();

   ostream             *outputStream_d;
   streambuf           *outputStreamBuf_d;
   int                  fShouldDeleteOutputStream_d;
   char                *pszDriverName_d;
   char                *pszDeviceName_d;
   char                *pszShortName_d;
   long                 lCapabilities_d;
   long                 lRasterCapabilities_d;
   long                 lDeviceOptions_d;
   long                 lDeviceOptions2_d;
   DeviceInstance      *pInstance_d;
   DeviceBlitter       *pBlitter_d;
   PDL                 *pPDL_d;

   char                *pszJobProperties_d;

   DeviceOrientation   *pOrientation_d;
   char                *pszDitherID_d;
   DeviceForm          *pForm_d;
   DeviceTray          *pTray_d;
   DeviceMedia         *pMedia_d;
   DeviceResolution    *pResolution_d;
   DeviceCommand       *pCommand_d;
   DevicePrintMode     *pPrintMode_d;
   DeviceData          *pData_d;

   char                *pszLoadedLibrary_d;  // @TBD expand to many
   GModule             *hModLibrary_d;

   PFNOUTPUTFUNCTION    pfnOutputFunction_d;
   void                *pMagicCookie_d;
};

#endif
