/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#include "DeviceInfo.hpp"
#include "OmniDomParser.hpp"

#include <sys/stat.h>

const static bool fDebugOutput = false;

extern void convertFilename (char *pszFileName);

char *vpszExeName = 0;

bool
areValidFiles (string *pstrFileName)
{
   string stringFileNameHPP (pstrFileName->c_str ());
   string stringFileNameCPP (pstrFileName->c_str ());

   convertFilename (const_cast<char*>(stringFileNameHPP.c_str ())); // @HACK?
   convertFilename (const_cast<char*>(stringFileNameCPP.c_str ())); // @HACK?

   stringFileNameHPP.append (".hpp");
   stringFileNameCPP.append (".cpp");

   struct stat statOutHPP;
   struct stat statOutCPP;

   if (-1 == stat (stringFileNameHPP.c_str (), &statOutHPP))
      return false;

   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameHPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutHPP.st_atime << endl;
      cerr        << "   mtime = " << statOutHPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutHPP.st_ctime << endl;
   }

   if (-1 == stat (stringFileNameCPP.c_str (), &statOutCPP))
      return false;

   if (fDebugOutput)
   {
      cerr << "stats for \"" << stringFileNameCPP << "\"" << endl;
      cerr << dec << "   atime = " << statOutCPP.st_atime << endl;
      cerr        << "   mtime = " << statOutCPP.st_mtime << endl;
      cerr        << "   ctime = " << statOutCPP.st_ctime << endl;
   }

   if (  0 == statOutHPP.st_size
      || 0 == statOutCPP.st_size
      )
      return false;
   else
      return true;
}

int
main (int argc, char *argv[])
{
   bool fErrorEncountered = false;

   vpszExeName = argv[0];

   char achLine[512]; // @TBD

   // Initialize the XML4C system
   try
   {
      XMLPlatformUtils::Initialize();
   }
   catch (const XMLException& toCatch)
   {
      cerr << "Error during initialization! :"
           << endl
           << XMLString::transcode (toCatch.getMessage ())
           << endl;

      return 1;
   }

   while (0 < cin.getline (achLine, sizeof (achLine)))
   {
//////cout << "line = \"" << achLine << "\"" << endl;

      if (  '#' != achLine[0]
         && sizeof (achLine) - 4 > strlen (achLine)
         )
      {
         DeviceInfo di;

         di.pstringCommandClassName     = 0;
         di.pstringOrientationClassName = 0;
         di.pstringResolutionClassName  = 0;
         di.pstringPrintModeClassName   = 0;
         di.pstringTrayClassName        = 0;
         di.pstringFormClassName        = 0;
         di.pstringMediaClassName       = 0;
         di.pstringGammaClassName       = 0;
         di.pstringDataClassName        = 0;
         di.pstringInstanceClassName    = 0;
         di.pstringBlitterClassName     = 0;
         di.pstringBlitterLibrary       = 0;
         di.pstringBlitterType          = 0;

         strcat (achLine, ".xml");

         OmniDomParser parser (achLine, &di);

         if (parser.parse ())
         {
            if (!areValidFiles (di.pstringCommandClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringCommandClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringOrientationClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringOrientationClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringResolutionClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringResolutionClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringPrintModeClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringPrintModeClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringTrayClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringTrayClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringFormClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringFormClassName << "*" << endl;
            }
            if (!areValidFiles (di.pstringMediaClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringMediaClassName << "*" << endl;
            }
////////////if (!areValidFiles (di.pstringGammaClassName))
////////////{
////////////   fErrorEncountered = true;
////////////   cerr << "Error: Could not create " << *di.pstringGammaClassName << "*" << endl;
////////////}
            if (!areValidFiles (di.pstringInstanceClassName))
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringInstanceClassName << "*" << endl;
            }
            if (  di.pstringBlitterClassName
               && !areValidFiles (di.pstringBlitterClassName)
               )
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringBlitterClassName << "*" << endl;
            }
            if (  di.pstringDataClassName
               && !areValidFiles (di.pstringDataClassName)
               )
            {
               fErrorEncountered = true;
               cerr << "Error: Could not create " << *di.pstringDataClassName << "*" << endl;
            }
         }
         else
         {
            fErrorEncountered = true;
         }

         // Clean up!
         delete di.pstringCommandClassName;
         delete di.pstringOrientationClassName;
         delete di.pstringResolutionClassName;
         delete di.pstringPrintModeClassName;
         delete di.pstringTrayClassName;
         delete di.pstringFormClassName;
         delete di.pstringMediaClassName;
         delete di.pstringGammaClassName;
         delete di.pstringDataClassName;
         delete di.pstringInstanceClassName;
         delete di.pstringBlitterClassName;
         delete di.pstringBlitterLibrary;
         delete di.pstringBlitterType;
      }
   }

   // And call the termination method
   XMLPlatformUtils::Terminate ();

   return fErrorEncountered;
}
