/*
 *   IBM Omni driver
 *   Copyright (c) International Business Machines Corp., 2000
 *
 *   This library is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as published
 *   by the Free Software Foundation; either version 2.1 of the License, or
 *   (at your option) any later version.
 *
 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
 *   the GNU Lesser General Public License for more details.
 *
 *   You should have received a copy of the GNU Lesser General Public License
 *   along with this library; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
#ifndef _OmniDomParser
#define _OmniDomParser

#include <string>

#include <util/PlatformUtils.hpp>
#include <parsers/DOMParser.hpp>
#include <dom/DOM_Node.hpp>

#include "MyErrorHandler.hpp"

#include "Device.hpp"
#include "DeviceInfo.hpp"

class OmniDomParser : public ReportError
{
public:
            OmniDomParser (const char *pszXMLFile, DeviceInfo *pdi);
   virtual ~OmniDomParser ();

   bool    parse                    ();
   void    setErrorCondition        (); // for ReportError interface

private:
   string *getNameOut                ();
   bool    shouldCreateFile          (string       strFileName);
   void    copyFile                  (const char  *pszFileNameIn);
   bool    parseBinaryData           (const char  *pszData,
                                      byte       **ppbData,
                                      int         *pcbData);
   void    processDOMNode            (DOM_Node&    node);
   bool    processDeviceOrientations (DOM_Node     deviceOrientationsNode);
   bool    processDeviceCommands     (DOM_Node     deviceCommandNode);
   bool    processDeviceResolutions  (DOM_Node     deviceResolutionNode);
   bool    processDevicePrintModes   (DOM_Node     devicePrintModes);
   bool    processDeviceTrays        (DOM_Node     deviceTrays);
   bool    processDeviceForms        (DOM_Node     deviceForms);
   bool    processDeviceMedias       (DOM_Node     deviceMedias);
   bool    processDeviceConnections  (DOM_Node     deviceConnections);
   bool    processDeviceGammas       (DOM_Node     deviceGammas);
   bool    processDeviceDatas        (DOM_Node     deviceDatas);
   bool    preprocessDriver          (DOM_Node     driverNode);
   bool    processDriver             (DOM_Node     driverNode);

   DOMParser  *parser_d;
   string     *pstrXMLFile_d;
   DeviceInfo *pdi_d;
   bool        fErrorCondition_d;
};

#endif
